/*
 * This Cplant(TM) source code is part of the Portals3 Reference
 * Implementation.
 *
 * This Cplant(TM) source code is the property of Sandia National
 * Laboratories.
 *
 * This Cplant(TM) source code is copyrighted by Sandia National
 * Laboratories.
 *
 * The redistribution of this Cplant(TM) source code is subject to the
 * terms of version 2 of the GNU General Public License.
 * (See COPYING, or http://www.gnu.org/licenses/lgpl.html.)
 *
 * Cplant(TM) Copyright 1998-2006 Sandia Corporation. 
 *
 * Under the terms of Contract DE-AC04-94AL85000, there is a non-exclusive
 * license for use of this work by or on behalf of the US Government.
 * Export of this program may require a license from the United States
 * Government.
 */


/* Portals3 is free software; you can redistribute it and/or modify
 * it under the terms of version 2 of the GNU General Public License,
 * as published by the Free Software Foundation.
 *
 * Portals3 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Portals3; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Questions or comments about this library should be sent to:
 *
 * Jim Schutt
 * Sandia National Laboratories, New Mexico
 * P.O. Box 5800
 * Albuquerque, NM 87185-0806
 *
 * jaschut@sandia.gov
 *
 */

#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/module.h>
#include <asm/siginfo.h>
#include <asm/signal.h>

/* These are all Portals3 include files.
 */
#include "p3utils.h"

/* Some of the NALs have a progress thread that uses select() to wake
 * up when there is data to move.  When such NALs have a new request
 * to send, the fd for the socket may not currently be in the write
 * fd_set if no data is currently being sent on that socket. 
 *
 * Such NALs don't want to use a finite timeout in select() to exit
 * and update the write fd_set, as this will either introduce a higher
 * load (in the short timeout, low latency case) or a longer latency.
 *
 * Instead such NALs want use an infinite select() timeout, and send a
 * signal (we use SIGSTOP) to the progress thread to interrupt
 * select(), so the write fd_set can be updated with the fd for the
 * socket with data to send.  This results in the lowest latency on
 * new sends.
 *
 * The downside to this design is we never know which of our system
 * calls might get interrupted by the signal, and to clear it we need to
 * pull the pending signal off the queue.  Since the progress thread is a
 * kernel thread, this doesn't happen as part of the signal delivery.
 *
 * So, use this function any time there is an error return code that might
 * signify a waiting signal, to dequeue the signal.  It takes the error
 * code from the syscall, and massages it if needed to return an error
 * code as would be seen in user-space.
 */
int p3_dequeue_signal(int err)
{
	siginfo_t info;
	unsigned long signr;

	if (signal_pending(current)) {

		spin_lock_irq(&current->sighand->siglock);
		signr = dequeue_signal(current, &current->blocked, &info);
		spin_unlock_irq(&current->sighand->siglock);

		/* Check if this is the signal we sent ourselves to wake up;
		 * maybe massage error code.
		 */
		if (signr == SIGSTOP) {

/* FIXME: need a p3 debug value writable in /sys so we can turn on
 * debugging.
			p3_print("p3_dequeue_signal: process %p "
				 "dequeued signal %ld\n", current, signr);
 */
			switch (err) {
			case -ERESTARTNOHAND:
				err = -EINTR;
				break;
			default:
				break;
			}
		}
		else
			p3_print("p3_dequeue_signal: unexpected signal %ld\n",
				 signr);
	}
	return err;
}

EXPORT_SYMBOL_GPL(p3_dequeue_signal);
