/*
 * This Cplant(TM) source code is part of the Portals3 Reference
 * Implementation.
 *
 * This Cplant(TM) source code is the property of Sandia National
 * Laboratories.
 *
 * This Cplant(TM) source code is copyrighted by Sandia National
 * Laboratories.
 *
 * The redistribution of this Cplant(TM) source code is subject to the
 * terms of version 2 of the GNU General Public License.
 * (See COPYING, or http://www.gnu.org/licenses/lgpl.html.)
 *
 * Cplant(TM) Copyright 1998-2006 Sandia Corporation. 
 *
 * Under the terms of Contract DE-AC04-94AL85000, there is a non-exclusive
 * license for use of this work by or on behalf of the US Government.
 * Export of this program may require a license from the United States
 * Government.
 */


/* Portals3 is free software; you can redistribute it and/or modify
 * it under the terms of version 2 of the GNU General Public License,
 * as published by the Free Software Foundation.
 *
 * Portals3 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Portals3; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Questions or comments about this library should be sent to:
 *
 * Jim Schutt
 * Sandia National Laboratories, New Mexico
 * P.O. Box 5800
 * Albuquerque, NM 87185-0806
 *
 * jaschut@sandia.gov
 *
 */
#include <p3-config.h>

#include "stdint.h"

#include <linux/list.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/err.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/device.h>
#include <linux/cdev.h>
#include <asm/uaccess.h>

/* These are all Portals3 include files.
 */
#include "p3utils.h"

#include <p3api/types.h>
#include <p3api/debug.h>

#include <p3/lock.h>
#include <p3/handle.h>
#include <p3/process.h>
#include <p3/forward.h>
#include <p3/errno.h>
#include <p3/debug.h>

#include <p3lib/types.h>
#include <p3lib/p3lib.h>
#include <p3lib/nal.h>
#include <p3lib/p3lib_support.h>
#include <p3lib/dispatch.h>

#include <p3/obj_alloc.h>	/* MUST come after p3lib/p3lib.h */

#include <p3rt/p3rt.h>
#include "p3_module.h"
#include "p3lib_module.h"

EXPORT_SYMBOL_GPL(p3lib_debug);	/* from p3lib/debug.h */

unsigned int dbg;
module_param(dbg, uint, S_IRUGO | S_IWUSR);

#define P3LIB_VERSION_STRING "Portals Message Passing Library v"PACKAGE_VERSION

struct list_head p3proc_pid_hash[P3_PROCESS_HASH_BINS];
struct list_head p3proc_tsk_hash[P3_PROCESS_HASH_BINS];

int p3lib_open(struct inode *inode, struct file *filep)
{
	p3_process_t *pp;

	if (DEBUG_P3(p3lib_debug, PTL_DBG_SETUP))
		p3_print("p3lib_open\n");

	pp = p3lib_new_process();
	if (!pp)
		return -ENOMEM;

	filep->private_data = pp;
	return 0;
}

static
void whack_ni_all(p3_process_t *pp)
{
	int i;

	if (pp->init < 0)
		return;

	for (i=0; i<PTL_MAX_INTERFACES; i++) {
		if (pp->ni[i])
			lib_PtlNIFini(pp->ni[i], NULL, NULL);
	}
}

int p3lib_flush(struct file *filep)
{
	if (DEBUG_P3(p3lib_debug, PTL_DBG_SHUTDOWN))
		p3_print("p3lib_flush\n");

	return 0;
}

int p3lib_release(struct inode *inode, struct file *filep)
{
	p3_process_t *pp = filep->private_data;

	if (DEBUG_P3(p3lib_debug, PTL_DBG_SETUP))
		p3_print("p3lib_release: p3_process %p\n", pp);

	if (!pp)
		return 0;

	whack_ni_all(pp);
	p3lib_free_process(pp);
	list_del(&pp->tsk_list);
	p3_free(pp);

	filep->private_data = NULL;

	return 0;
}

/* p3lib_write implements the kernel side of Portals request forwarding.
 * It expects to read a buffer containing two (pointer, length) pairs,
 * where the first describes the request block, and the second describes
 * the result block.
 */
ssize_t p3lib_write(struct file *filep,
		    const char __user *buf, size_t count, loff_t *f_pos)
{
	p3_forward_t fw;
	p3_libreq_t *req;
	p3_libres_t *res;
	ssize_t ret = sizeof(fw);

	*f_pos = 0;

	if (count != sizeof(fw)) {
		ret = -EINVAL;
		goto fail;
	}
	if (copy_from_user(&fw, buf, sizeof(fw))) {
		ret = -EFAULT;
		goto fail;
	}
	if (!(req = p3_malloc(fw.req_len))) {
		ret = -ENOMEM;
		goto fail;
	}
	if (!(res = p3_malloc(fw.res_len))) {
		ret = -ENOMEM;
		goto fail_req;
	}
	if (copy_from_user(req, (char *)fw.request, fw.req_len)) {
		ret = -EFAULT;
		goto fail_res;
	}
	if (copy_from_user(res, (char *)fw.result, fw.res_len)) {
		ret = -EFAULT;
		goto fail_res;
	}
	res->status = PTL_FAIL;
	lib_dispatch(req, res);
	if (copy_to_user((char *)fw.result, res, fw.res_len))
		ret = -EFAULT;

fail_res:
	p3_free(res);
fail_req:
	p3_free(req);
fail:
	return ret;
}

struct file_operations p3lib_fops = {
	.owner = THIS_MODULE,
	.open = p3lib_open,
	.flush = p3lib_flush,
	.release = p3lib_release,
	.write = p3lib_write,
};

static
ssize_t attr_p3lib_dev_show(struct class_device *cls, char *buf)
{
	return snprintf(buf, PAGE_SIZE, "%d:%d\n", P3LIB_MAJOR, P3LIB_MINOR);
}

static
void p3lib_classdev_release(struct class_device *dev)
{
	/* We don't have an object which imbeds a struct class_device, so
	 * we have nothing to clean up or free.
	 */
}

dev_t p3_dev;
EXPORT_SYMBOL_GPL(p3_dev);

static  struct cdev p3lib_cdev;

struct class p3_class = {
	.name = "portals3",
	.release = p3lib_classdev_release,
};
EXPORT_SYMBOL_GPL(p3_class);

static
struct class_device p3lib_class_dev = {
	.class = &p3_class,
};

static
struct class_device_attribute attr_p3lib_dev = {
	.attr = { .name = "dev",
		  .owner = THIS_MODULE,
		  .mode = S_IRUGO
	},
	.show = attr_p3lib_dev_show,
};

static
int __init p3lib_init(void)
{
	int i, err;

	if (dbg != p3lib_debug)
		p3lib_debug = dbg;

	for (i=0; i<P3_PROCESS_HASH_BINS; i++) {
		INIT_LIST_HEAD(p3proc_pid_hash+i);
		INIT_LIST_HEAD(p3proc_tsk_hash+i);
	}
	err = alloc_chrdev_region(&p3_dev,
				  P3LIB_MINOR, P3_NUM_MINORS, "portals3");
	if (err < 0) {
		printk(KERN_WARNING "portals3: no char dev available.\n");
		goto fail;
	}
	err = class_register(&p3_class);
	if (err < 0)
		goto fail_dev;

	strlcpy(p3lib_class_dev.class_id, "p3lib", BUS_ID_SIZE);
	err = class_device_register(&p3lib_class_dev);
	if (err) 
		goto fail_class;

	err = class_device_create_file(&p3lib_class_dev, &attr_p3lib_dev);
	if (err)
		goto fail_class_device;

	/* Turn on the device.  Last.
	 */
	cdev_init(&p3lib_cdev, &p3lib_fops);
	p3lib_cdev.owner = THIS_MODULE;
	err = cdev_add(&p3lib_cdev, P3LIB_DEV, 1);
	if (err)
		goto fail_class_dev_attr;

	printk(KERN_INFO P3LIB_VERSION_STRING"\n");
	return 0;

fail_class_dev_attr:
	class_device_remove_file(&p3lib_class_dev, &attr_p3lib_dev);
fail_class_device:
	class_device_unregister(&p3lib_class_dev);
fail_class:
	class_unregister(&p3_class);
fail_dev:
	unregister_chrdev_region(p3_dev, P3_NUM_MINORS);
fail:
	return err;
}

static
void __exit p3lib_exit(void)
{
	cdev_del(&p3lib_cdev);
	class_device_remove_file(&p3lib_class_dev, &attr_p3lib_dev);

	if (DEBUG_P3(p3lib_debug, PTL_DBG_SHUTDOWN))
		p3_print("p3lib_exit: removing sysfs class device %s/%s\n",
			 p3_class.name, p3lib_class_dev.class_id);
	class_device_unregister(&p3lib_class_dev);

	if (DEBUG_P3(p3lib_debug, PTL_DBG_SHUTDOWN))
		p3_print("p3lib_exit: removing sysfs class %s\n",
			 p3_class.name);
	class_unregister(&p3_class);
	unregister_chrdev_region(p3_dev, P3_NUM_MINORS);
}

module_init(p3lib_init);
module_exit(p3lib_exit);

MODULE_LICENSE("GPL v2");
MODULE_AUTHOR("Jim Schutt");
MODULE_VERSION(PACKAGE_VERSION);
MODULE_DESCRIPTION(P3LIB_VERSION_STRING);

/* For symbols that have the same implementation is kernel and user
 * space, We'll export all those we need from this module here, rather
 * than exporting them in their source files.
 */
EXPORT_SYMBOL_GPL(lib_update);
EXPORT_SYMBOL_GPL(lib_register_nal);
EXPORT_SYMBOL_GPL(lib_unregister_nal);
EXPORT_SYMBOL_GPL(lib_parse);
EXPORT_SYMBOL_GPL(lib_finalize);
EXPORT_SYMBOL_GPL(runtime_req_pid);
