/*
 * This Cplant(TM) source code is part of the Portals3 Reference
 * Implementation.
 *
 * This Cplant(TM) source code is the property of Sandia National
 * Laboratories.
 *
 * This Cplant(TM) source code is copyrighted by Sandia National
 * Laboratories.
 *
 * The redistribution of this Cplant(TM) source code is subject to the
 * terms of version 2 of the GNU General Public License.
 * (See COPYING, or http://www.gnu.org/licenses/lgpl.html.)
 *
 * Cplant(TM) Copyright 1998-2006 Sandia Corporation. 
 *
 * Under the terms of Contract DE-AC04-94AL85000, there is a non-exclusive
 * license for use of this work by or on behalf of the US Government.
 * Export of this program may require a license from the United States
 * Government.
 */


/* Portals3 is free software; you can redistribute it and/or modify
 * it under the terms of version 2 of the GNU General Public License,
 * as published by the Free Software Foundation.
 *
 * Portals3 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Portals3; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Questions or comments about this library should be sent to:
 *
 * Jim Schutt
 * Sandia National Laboratories, New Mexico
 * P.O. Box 5800
 * Albuquerque, NM 87185-0806
 *
 * jaschut@sandia.gov
 *
 */
#include <p3-config.h>

/* This file is used to implement Portals library support functions that
 * are specific to a kernel-space library implementation.
 */

#include <linux/list.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/module.h>
#include <linux/highmem.h>
#include <asm/scatterlist.h>

/* These are all Portals3 include files.
 */
#include "p3utils.h"

#include <p3api/types.h>
#include <p3api/misc.h>
#include <p3api/debug.h>

#include <p3/lock.h>
#include <p3/handle.h>
#include <p3/process.h>
#include <p3/forward.h>
#include <p3/nal_types.h>
#include <p3/errno.h>
#include <p3/debug.h>

#include <p3lib/types.h>
#include <p3lib/p3lib.h>
#include <p3lib/nal.h>
#include <p3lib/p3lib_support.h>
#include <p3lib/p3validate.h>

#include <p3/obj_alloc.h>	/* MUST come after p3lib/p3lib.h */

#include "p3lib_module.h"

/* Call with lib_update lock held.
 */
static inline
p3_process_t *__get_p3p_task(void *task_ptr)
{
	p3_process_t *pp;
	struct list_head *head = p3proc_tsk_listhead(task_ptr);

	if (!list_empty(head)) {
		list_for_each_entry(pp, head, tsk_list) {
			if (pp->user_task == task_ptr)
				return pp;
		}
	}
	return NULL;
}

/* Call with lib_update lock held.
 */
static inline
lib_ni_t *__get_ni_pid(ptl_interface_t type, ptl_pid_t pid)
{
	lib_ni_t *ni;
	struct list_head *head = p3proc_pid_listhead(pid);

	if (!list_empty(head)) {
		list_for_each_entry(ni, head, pid_list) {
			if (ni->nal->nal_type->type == type && ni->pid == pid)
				return ni;
		}
	}
	return NULL;
}

/* Call with lib_update lock held.
 */
static inline
p3_process_t *__get_p3p_pid(ptl_interface_t type, ptl_pid_t pid)
{
	lib_ni_t *ni = __get_ni_pid(type, pid);

	if (ni)
		return ni->owner;
	return NULL;
}

/* Since p3lib_fops.p3lib_open allocated the portals process struct, we
 * cannot free it here - p3lib_flush will do so.  But, we return it
 * to the uninitialized state here.
 */
void p3lib_free_process(p3_process_t *pp)
{
	unsigned i;

	if (!pp)
		return;

	p3_lock(&lib_update);

	for (i=0; i<PTL_MAX_INTERFACES; i++) {
		if (pp->ni[i]) {
			p3_unlock(&lib_update);
			PTL_ROAD();
		}
	}
	if (pp->next_group) {
		p3_free(pp->next_group);
		pp->next_group = NULL;
	}
	pp->jid = PTL_JID_ANY;
	pp->init = -1;

	p3_unlock(&lib_update);
}

p3_process_t *__p3lib_new_process(void)
{
	p3_process_t *pp;

	if (DEBUG_P3(p3lib_debug, PTL_DBG_MEMORY))
		p3_print("__p3lib_new_process\n");

	if (!(pp = p3_malloc(sizeof *pp)))
		return NULL;
	memset(pp, 0, sizeof(p3_process_t));

	pp->jid = PTL_JID_ANY;
	pp->user_task = current;
	pp->debug = p3lib_debug;

	/* Use the Portals process object's init value to track the number of
	 * interfaces used by the process, or -1 to indicate PtlInit not
	 * called.
	 */
	pp->init = -1;

	p3_lock(&lib_update);
	list_add(&pp->tsk_list, p3proc_tsk_listhead(current));
	p3_unlock(&lib_update);

	return pp;
}

p3_process_t *p3lib_new_process(void)
{
	p3_process_t *pp;

	p3_lock(&lib_update);
	pp = __get_p3p_task(current);
	p3_unlock(&lib_update);

	if (!pp) 
		pp = __p3lib_new_process();

	return pp;
}

p3_process_t *__p3lib_cur_process(void)
{
	p3_process_t *pp = __get_p3p_task(current);
	if (pp && pp->init < 0)
		pp = NULL;
	return pp;
}
EXPORT_SYMBOL_GPL(__p3lib_cur_process);

p3_process_t *p3lib_cur_process(void)
{
	p3_process_t *pp;

	p3_lock(&lib_update);
	pp = __p3lib_cur_process();
	p3_unlock(&lib_update);

	return pp;
}
EXPORT_SYMBOL_GPL(p3lib_cur_process);

p3_process_t *p3lib_get_process(ptl_interface_t type, ptl_pid_t pid)
{
	p3_process_t *pp;

	p3_lock(&lib_update);
	pp = __get_p3p_pid(type, pid);
	p3_unlock(&lib_update);

	return pp;
}

lib_ni_t *p3lib_get_ni_pid(ptl_interface_t type, ptl_pid_t pid)
{
	lib_ni_t *ni = NULL;

	p3_lock(&lib_update);
	ni = __get_ni_pid(type, pid);
	p3_unlock(&lib_update);

	return ni;
}
EXPORT_SYMBOL_GPL(p3lib_get_ni_pid);

/* Only safe to call when holding lib_update lock.
 */
void __p3lib_process_add_pid(lib_ni_t *ni, ptl_pid_t pid)
{
	ni->pid = pid;
	list_add(&ni->pid_list, p3proc_pid_listhead(pid));
}

/* Only safe to call when holding lib_update lock.
 */
void __p3lib_process_rel_pid(lib_ni_t *ni)
{
	list_del_init(&ni->pid_list);
	ni->pid = PTL_PID_ANY;
}

/* We don't need any setup/teardown, and NAL modules will register themselves
 * as they are inserted.
 */
void p3lib_nal_setup(void) {}
void p3lib_nal_teardown(void) {}

#ifndef NO_PINNED_PAGES

#ifdef PORTALS_PROFILE
#define profile_uaddr_align(addr,ni)	\
do {							\
	if (!((addr) & 0x7fUL))				\
		ni_stat_inc(ni,PTL_SR_MD_ALIGN_128,1);	\
	else if (!((addr) & 0x3fUL))			\
		ni_stat_inc(ni,PTL_SR_MD_ALIGN_64,1);	\
	else if (!((addr) & 0x1fUL))			\
		ni_stat_inc(ni,PTL_SR_MD_ALIGN_32,1);	\
	else if (!((addr) & 0xfUL))			\
		ni_stat_inc(ni,PTL_SR_MD_ALIGN_16,1);	\
	else						\
		ni_stat_inc(ni,PTL_SR_MD_ALIGN_8,1);	\
} while 0
#else
#define profile_uaddr_align(addr,ni) do {} while (0)
#endif

static inline
void free_p3_addrmap(struct p3_addrmap *am)
{
	unsigned i, j;

	if (!am)
		return;

	for (i=0; i<am->maplen; i++)
		if (am->map[i]) {
			struct p3_ubuf_map *ub_map = am->map[i];

			for (j=0; j<ub_map->sglen; j++)
				kfree(ub_map->sglist[j]);

			kfree(ub_map);
		}
		else
			break;
	kfree(am);
}

int p3lib_vvalidate(lib_ni_t *ni,
		    ptl_md_iovec_t *iov, size_t iovlen, void **addrkey)
{
	struct p3_addrmap *am;
	unsigned i, j, sz, np, tnp = 0, empty = 1;
	int rc = PTL_OK;

	if (DEBUG_P3(ni->debug, PTL_DBG_MEMORY))
		p3_print("p3lib_vvalidate: len %u iov@ %p ptr_to_key@ %p\n",
			 (unsigned)iovlen, iov, addrkey);

	if (!addrkey)
		goto out;
	if (!(iov && iovlen)) {
		*addrkey = NULL;
		goto out;
	}
	sz = sizeof(struct p3_addrmap)
		+ (iovlen-1) * sizeof(struct p3_ubuf_map *);
	am = kmalloc(sz, GFP_KERNEL);
	if (!am) {
		rc = PTL_NO_SPACE;
		goto out;
	}
	memset(am, 0, sz);
	INIT_LIST_HEAD(&am->msglist);
	am->wait = 0;
	am->maplen = iovlen;
	am->task = current;

	for (i=0; i<iovlen; i++) {

		struct p3_ubuf_map *ub_map;
		unsigned long base = (unsigned long)iov[i].iov_base;

		if (DEBUG_P3(ni->debug, PTL_DBG_MEMORY))
			p3_print("p3lib_vvalidate: %u bytes@ %p "
				 "ptr_to_key@ %p\n", (unsigned)iov[i].iov_len,
				 (void*)base, addrkey);

		profile_uaddr_align(base, ni);

		np = ((base + iov[i].iov_len) >> PAGE_SHIFT) -
			(base >> PAGE_SHIFT) + 1;

		sz = sizeof(struct p3_ubuf_map)
			+ (np-1) * sizeof(struct scatterlist*);
		
		ub_map = kmalloc(sz, GFP_KERNEL);
		if (!ub_map) {
			rc = PTL_NO_SPACE;
			goto out_free;
		}
		memset(ub_map, 0, sz);
		ub_map->uaddr = base;
		ub_map->length = iov[i].iov_len;
		ub_map->sglen = np;

		/* Here we're trying to avoid high-order allocations,
		 * which would be needed if we kmalloced the whole array
		 * at once.
		 */
		for (j=0; j<np; j++) {
			ub_map->sglist[j] = kmalloc(sizeof(struct scatterlist),
							   GFP_KERNEL);
			if (!ub_map->sglist[j]) {
				rc = PTL_NO_SPACE;
				goto out_free;
			}
			memset(ub_map->sglist[j], 0, 
			       sizeof(struct scatterlist));
		}
		am->map[i] = ub_map;

		tnp += np;
		if (iov[i].iov_len)
			empty = 0;
	}
	if (empty) {
		*addrkey = NULL;
		goto out_free;
	}
	rc = p3validate(am);
	if (rc != PTL_OK)
		goto out_free;

	if (DEBUG_P3(ni->debug, PTL_DBG_MEMORY))
		p3_print("p3lib_vvalidate: mapped %u pages, key@ %p\n",
			 tnp, am);
	*addrkey = am;
out:
	return rc;
out_free:
	free_p3_addrmap(am);
	goto out;
}

#else /* NO_PINNED_PAGES */

int p3lib_vvalidate(lib_ni_t *ni,
		    ptl_md_iovec_t *iov, size_t iovlen, void **addrkey)
{
	if (addrkey)
		*addrkey = P3_MSG_ADDRMAP_NO_PIN;

	return PTL_OK;
}

#endif /* NO_PINNED_PAGES */

EXPORT_SYMBOL_GPL(p3lib_vvalidate);

int p3lib_validate(lib_ni_t *ni,
		   api_mem_t *base, size_t extent, void **addrkey)
{
	ptl_md_iovec_t iov;

	if (!addrkey) return 0;
	if (!extent) {
		*addrkey = NULL;
		return PTL_OK;
	}
	iov.iov_base = base;
	iov.iov_len = 1;

	return p3lib_vvalidate(ni, &iov, 1, addrkey);
}
EXPORT_SYMBOL_GPL(p3lib_validate);

#ifndef NO_PINNED_PAGES

/* Sanity p3lib_check for vinvalidate(); but PtlMDUpdate doesn't have to
 * revalidate if all the iovec entries have the same base addresses, but
 * extents are shorter.
 */
#ifdef DEBUG_PTL_INTERNALS
#define addrmap_sanity_check(iov,iovlen,am)	\
do {									\
	unsigned i;							\
									\
	if (iovlen != am->maplen)					\
		PTL_ROAD();						\
									\
	for (i=0; i<iovlen; i++) {					\
		if (iov[i].iov_base != (void *)am->map[i]->uaddr ||	\
		    iov[i].iov_len > am->map[i]->length)		\
			PTL_ROAD();					\
	}								\
} while (0)
#else
#define addrmap_sanity_check(iov,iovlen,am) do {} while (0)
#endif

void p3lib_vinvalidate(lib_ni_t *ni,
		       ptl_md_iovec_t *iov, size_t iovlen, void *addrkey)
{
	struct p3_addrmap *am = addrkey;
	struct list_head *item;

	if (!am)
		return;

	/* Normal flow of execution is that memory for an MD is invalidated 
	 * when the MD is unlinked, usually because a message was completed.
	 *
	 * However, under exceptional conditions, such as when a program 
	 * terminates abnormally, memory for an MD may be invalidated at any
	 * time, including when it is still being used to service an active
	 * message.  In this case, we need to wait until the data movement
	 * completes before we invalidate, or we risk scribbling over
	 * something important if the pages get reused immediately.
	 *
	 * In the normal case, am->wait == 0 && list_empty(&am->msglist)
	 * always, and we can just invalidate the memory and free the addrmap.
	 *
	 * In the exceptional case, we delete any message from am->msglist
	 * that is not at the head of a NAL message queue.  If any messages
	 * are still referencing our addrmap, we go to sleep, assuming that a
	 * NAL will continue to make progess on such messages, and they will
	 * complete.  When each such message completes, the NAL should drop
	 * its reference to our addrmap, wake us up if it sees we are sleeping
	 * on a message completion, and continue with finalizing the message.
	 *
	 * Finalizing the completed message may involve unlinking the MD used
	 * to process it, which will cause p3lib_invalidate() to be called.
	 * At which point we have a race between two threads trying to
	 * invalidate the same memory.
	 *
	 * Hence am->wait.  It gets set when the exceptional case needs to
	 * sleep on a message completion (call this thread A).  The completion
	 * thread (call it thread B) finishes a message, drops its reference
	 * to the addrmap, wakes up A, and then ends up here by calling
	 * lib_finalize(), where it finds am->wait == 1.  B busy-waits to make
	 * sure A has woken and cleaned up from waiting, then finishes the
	 * invalidation clean-up.
	 *
	 * Simple, no?
	 */
	if (am->wait) {
		if (DEBUG_P3(ni->debug, PTL_DBG_MEMORY))
			p3_print("p3lib_vinvalidate: waiting on msg\n");
		while (am->wait)
			p3_yield();
		p3invalidate(am);
		free_p3_addrmap(am);
		return;
	}
	am->wait = 1;

	if (DEBUG_P3(ni->debug, PTL_DBG_MEMORY))
		p3_print("p3lib_vinvalidate: len %u iov@ %p ptr_to_key@ %p\n",
			 (unsigned)iovlen, iov, addrkey);

	addrmap_sanity_check(iov, iovlen, am);

	item = am->msglist.next;
	while (item != &am->msglist) {

		void *nal_int_msg;
		struct p3_addrmap_hook *hook;

		hook = list_entry(item, typeof(*hook), msglist_link);
		item = item->next;

		if (hook->msg_addrmap != am) {
			p3_print("p3lib_vinvalidate: hook->msg_addrmap %p "
				 "addrkey %p\n", hook->msg_addrmap, am);
			PTL_ROAD();
		}
		nal_int_msg = ni->nal->msg_container_of(ni, hook);

		if (DEBUG_P3(ni->debug, PTL_DBG_MEMORY))
			p3_print("p3lib_vinvalidate: dequeueing msg %p using "
				 "key@ %p\n", nal_int_msg, am);

		ni->nal->dequeue_msg(ni, nal_int_msg);

	}
	if (list_empty(&am->msglist)) {
		if (DEBUG_P3(ni->debug, PTL_DBG_MEMORY))
			p3_print("p3lib_vinvalidate: unmapping memory "
				 "using key@ %p\n", am);
		p3invalidate(am);
		free_p3_addrmap(am);
		return;
	}
	/* OK, there's at least one message using this map that that NAL 
	 * cannot release.
	 *
	 * We should only get here if this invalidate was requested under
	 * exceptional conditions, so we'll go to sleep waiting for the
	 * messages using this map to finish.  We must let the messages 
	 * finish for those NALs that use a stream-based implementation;
	 * otherwise we'd cause the NAL to loose track of where it was in
	 * the stream.
	 */
	{
		struct p3_addrmap_hook
			*hook = list_entry(am->msglist.next,
					   typeof(*hook), msglist_link);

		DECLARE_WAITQUEUE(wait, current);
		add_wait_queue(&hook->msg_waitq, &wait);

		if (DEBUG_P3(ni->debug, PTL_DBG_MEMORY)) {
			void *nal_int_msg;
			nal_int_msg = ni->nal->msg_container_of(ni, hook);
			p3_print("p3lib_vinvalidate: wait on NAL msg %p\n",
				 nal_int_msg);
		}

		set_current_state(TASK_INTERRUPTIBLE);
		while(hook->msg_addrmap)
			schedule();

		set_current_state(TASK_RUNNING);
		remove_wait_queue(&hook->msg_waitq, &wait);

		/* FIXME:  We'e going to deadlock if there are active messages
		 * all using this address map at the time the program is
		 * terminated abnormally.
		 *
		 * Instead, we'll oops so it gets fixed.
		 */
		if (!list_empty(&am->msglist))
			PTL_ROAD();

		am->wait = 0;
	}
}
#else /* NO_PINNED_PAGES */

void p3lib_vinvalidate(lib_ni_t *ni,
		       ptl_md_iovec_t *iov, size_t iovlen, void *addrkey) {}

#endif /* NO_PINNED_PAGES */

EXPORT_SYMBOL_GPL(p3lib_vinvalidate);

void p3lib_invalidate(lib_ni_t *ni,
		      api_mem_t *base, size_t extent, void *addrkey)
{
	ptl_md_iovec_t iov;

	if (!addrkey)
		return;

	iov.iov_base = base;
	iov.iov_len = 1;

	p3lib_vinvalidate(ni, &iov, 1, addrkey);
}
EXPORT_SYMBOL_GPL(p3lib_invalidate);


/* Lifted from i386 __copy_to_user_ll(), which is another reason this Portals
 * implementation has to be GPL.
 */
int p3lib_copy_to_api(lib_ni_t *ni,
		      api_mem_t *dst, void *src, ptl_size_t len)
{
	struct mm_struct *mm;
	struct task_struct *task = ni->owner->user_task;
	unsigned long l = len;

	if (l != len)	/* len is 64 bits.  is long? */
		return PTL_FAIL;

	/* The API might be in kernel space.
	 */
	if ((unsigned long)dst >= TASK_SIZE) {
		memcpy(dst, src, len);
		return PTL_OK;
	}

	mm = get_task_mm(task);
	if (!mm)
		return PTL_SEGV;

	while (len) {
		void *dst_kva;
		struct page *pg;
		unsigned long os = ((unsigned long)dst) & (PAGE_SIZE-1);

		l = PAGE_SIZE - os;
		if (l > len)
			l = len;

		down_read(&mm->mmap_sem);
		if (1 != get_user_pages(task, mm, (unsigned long)dst,
					1, 1, 1, &pg, NULL)) {
			up_read(&mm->mmap_sem);
			mmput(mm);
			return PTL_SEGV;
		}
		dst_kva = kmap_atomic(pg, KM_USER1);
		memcpy(dst_kva + os, src, l);
		kunmap_atomic(pg, KM_USER1);

		flush_dcache_page(pg);
		set_page_dirty_lock(pg);
		put_page(pg);
		up_read(&mm->mmap_sem);

		src += l;
		dst += l;
		len -= l;
	}
	mmput(mm);
	return PTL_OK;
}
EXPORT_SYMBOL_GPL(p3lib_copy_to_api);
