/*
 * This Cplant(TM) source code is part of the Portals3 Reference
 * Implementation.
 *
 * This Cplant(TM) source code is the property of Sandia National
 * Laboratories.
 *
 * This Cplant(TM) source code is copyrighted by Sandia National
 * Laboratories.
 *
 * The redistribution of this Cplant(TM) source code is subject to the
 * terms of version 2 of the GNU General Public License.
 * (See COPYING, or http://www.gnu.org/licenses/lgpl.html.)
 *
 * Cplant(TM) Copyright 1998-2006 Sandia Corporation. 
 *
 * Under the terms of Contract DE-AC04-94AL85000, there is a non-exclusive
 * license for use of this work by or on behalf of the US Government.
 * Export of this program may require a license from the United States
 * Government.
 */


/* Portals3 is free software; you can redistribute it and/or modify
 * it under the terms of version 2 of the GNU General Public License,
 * as published by the Free Software Foundation.
 *
 * Portals3 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Portals3; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Questions or comments about this library should be sent to:
 *
 * Jim Schutt
 * Sandia National Laboratories, New Mexico
 * P.O. Box 5800
 * Albuquerque, NM 87185-0806
 *
 * jaschut@sandia.gov
 *
 */
#include <asm/system.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/sched.h>
#include <linux/in.h>

#ifndef _PTL3_KERN_P3UTILS_H_
#define _PTL3_KERN_P3UTILS_H_

/* Some errors are so egregious we can't possibly continue.  We have to roll
 * over and die, whether we're in user or kernel space.
 */
#if 0
#define PTL_ROAD()  panic("Impossible")
#else
#define PTL_ROAD() do { BUG(); panic("Impossible"); } while (0)
#endif

/* Use the kernel's type-checking min/max macros
 */
#ifdef MAX
#undef MAX
#endif
#define MAX(a,b) max(a,b)

#ifdef MIN
#undef MIN
#endif
#define MIN(a,b) min(a,b)

#define memory_barrier_start() do { mb(); } while (0)
#define memory_barrier_end()   do { mb(); } while (0)

/* For a kernel-space Portals library, the module initialization routine
 * is the place for any initialization.  The P3 library needs this entry
 * point for a user-space implementation.
 */
#define p3utils_init() do {} while (0)

/* For a kernel-space Portals library, we can pull the effective user id
 * right out of the current task-struct.  This is only safe to call from
 * process context!
 */
#define geteuid() (current->euid)

#define P3MSG_EMERG   KERN_EMERG
#define P3MSG_ALERT   KERN_ALERT
#define P3MSG_CRIT    KERN_CRIT
#define P3MSG_ERR     KERN_ERR
#define P3MSG_WARNING KERN_WARNING
#define P3MSG_NOTICE  KERN_NOTICE
#define P3MSG_INFO    KERN_INFO
#define P3MSG_DEBUG   KERN_DEBUG

/* Use this if you want to differentiate between user and progress-thread
 * context in debug printing.  (Hint: debugging is much easier if you do.)
 */
#define P3_PRINT_SHOWS_CURRENT

#ifdef P3_PRINT_SHOWS_CURRENT
#define P3_PRINT_BUFSZ 512
extern char p3_print_buf[NR_CPUS][P3_PRINT_BUFSZ];
extern spinlock_t p3_print_lock;
#define p3_print(args...)	\
do {								\
	char *buf;						\
	int n;							\
								\
	spin_lock_irq(&p3_print_lock);				\
	buf = p3_print_buf[smp_processor_id()];			\
	n = snprintf(buf, P3_PRINT_BUFSZ, "%p: ", current);	\
	snprintf(buf + n, P3_PRINT_BUFSZ - n, args);		\
	printk("%s", buf);					\
	spin_unlock_irq(&p3_print_lock);			\
} while (0)
#else
#define p3_print(args...) do { printk(args); } while (0)
#endif

#define p3_flush() do {} while (0)

/* For now, we'll use atomic allocation, since most of our allocations
 * will be small, and we do them from inside a spin lock so we can't
 * sleep.
 */
#define p3_free(ptr) kfree(ptr)
#if 1
#define p3_malloc(sz) kmalloc(sz,GFP_ATOMIC)
#else
#define p3_malloc(sz) \
({									\
	if (sz <= 32)							\
		p3_print("kmalloc %d bytes: %s:%s:%d\n",		\
			 (int)(sz), __FUNCTION__, __FILE__,__LINE__);	\
	kmalloc(sz,GFP_ATOMIC);						\
})
#endif

static inline 
void *p3_realloc(void* ptr, size_t old_sz, size_t new_sz)
{
        void *nptr;
	size_t csz;
	
	if (!ptr) 
		return p3_malloc(new_sz);
        if (new_sz == old_sz)
		return ptr;

        csz = new_sz < old_sz ? new_sz : old_sz;
        if (new_sz) {
                if (!(nptr = p3_malloc(new_sz)))
			return nptr;
                memcpy(nptr, ptr, csz);
        }
	else
		nptr = NULL;

        p3_free(ptr);
        return nptr;
}

static inline
char *p3_strdup(const char *ptr)
{
	char *dup = NULL;

	if (ptr && (dup = p3_malloc(strlen(ptr)+1)))
		strcpy(dup, ptr);

	return dup;
}

static inline
void p3_yield(void)
{
	schedule();
}

void *bsearch(const void *key, const void *base, size_t nmemb, size_t size,
	      int (*compare)(const void*, const void*));

#define INET_NTOA_BUF_SZ 16
char *inet_ntoa(struct in_addr in);

typedef struct {
	unsigned long expire;
	unsigned set;
} timeout_val_t;

static inline
void set_timeout(timeout_val_t *to, unsigned msec_delay)
{
	to->set = 1;
	to->expire = jiffies + HZ * (unsigned long)msec_delay / 1000;
}

static inline
void clear_timeout(timeout_val_t *to)
{
	to->set = 0;
}

static inline
int timeout_set(timeout_val_t *to)
{
	return to->set;
}

/* Returns 0 if timeout has not expired, none-zero if it has.
 * BY definition, an unset timeout is never expired.
 */
static inline
int test_timeout(timeout_val_t *to)
{
	return !to->set || time_before(jiffies, to->expire) ? 0 : 1;
}

#endif /* _PTL3_KERN_P3UTILS_H_ */
