/*
 * This Cplant(TM) source code is part of the Portals3 Reference
 * Implementation.
 *
 * This Cplant(TM) source code is the property of Sandia National
 * Laboratories.
 *
 * This Cplant(TM) source code is copyrighted by Sandia National
 * Laboratories.
 *
 * The redistribution of this Cplant(TM) source code is subject to the
 * terms of version 2 of the GNU General Public License.
 * (See COPYING, or http://www.gnu.org/licenses/lgpl.html.)
 *
 * Cplant(TM) Copyright 1998-2006 Sandia Corporation. 
 *
 * Under the terms of Contract DE-AC04-94AL85000, there is a non-exclusive
 * license for use of this work by or on behalf of the US Government.
 * Export of this program may require a license from the United States
 * Government.
 */


/* Portals3 is free software; you can redistribute it and/or modify
 * it under the terms of version 2 of the GNU General Public License,
 * as published by the Free Software Foundation.
 *
 * Portals3 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Portals3; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Questions or comments about this library should be sent to:
 *
 * Jim Schutt
 * Sandia National Laboratories, New Mexico
 * P.O. Box 5800
 * Albuquerque, NM 87185-0806
 *
 * jaschut@sandia.gov
 *
 */
#include <p3-config.h>

/* This file is used to implement Portals library support functions that
 * are specific to a kernel-space library implementation.
 */

#include <linux/module.h>
#include <linux/list.h>
#include <linux/mm.h>

#include "p3utils.h"

#include <p3api/types.h>
#include <p3api/debug.h>

#include <p3/errno.h>
#include <p3/debug.h>

#include <p3lib/debug.h>
#include <p3lib/p3validate.h>

#include "p3_module.h"

#ifndef NO_PINNED_PAGES

#ifdef DEBUG_PTL_INTERNALS
#define page_sanity_check(ubuf_map,name)	\
do {									\
	size_t j;							\
	int c;								\
									\
	for (j=0; j<ubuf_map->sglen; j++) {				\
		if (!ubuf_map->sglist[j]->page) {			\
			p3_print(#name": page %p "			\
				 "("FMT_SZ_T"/"FMT_SZ_T")\n",		\
				 ubuf_map->sglist[j]->page, j,		\
				 ubuf_map->sglen);			\
			continue;					\
		}							\
		c = page_count(ubuf_map->sglist[j]->page);		\
		if (c >= 2)						\
			continue;					\
		p3_print(#name": page %p ("FMT_SZ_T"/"FMT_SZ_T") "	\
			 "count %d buf start %p len "FMT_SZ_T"\n",	\
			 ubuf_map->sglist[j]->page, j, ubuf_map->sglen,	\
			 c, (void *)ubuf_map->uaddr, ubuf_map->length);	\
	}								\
} while (0)
#else
#define page_sanity_check(ubuf_map,name) do {} while (0)
#endif

int p3validate(struct p3_addrmap *am)
{
	struct task_struct *task = am->task;
	struct mm_struct *mm;
	struct page **pages = NULL;
	size_t i, j, n, np = 0, ppp;

	ppp = PAGE_SIZE / sizeof(void *);
	pages = kmalloc(ppp * sizeof(void *), GFP_KERNEL);
	if (!pages)
		return PTL_NO_SPACE;

	mm = get_task_mm(task);

	for (i=0; i<am->maplen; i++) {
		int rc;
		size_t l = 0;
		struct p3_ubuf_map *m = am->map[i];
		unsigned long uaddr = m->uaddr;

		if (DEBUG_P3(p3lib_debug, PTL_DBG_MEMORY))
			p3_print("p3validate: map %p uaddr %p "
				 "len "FMT_SZ_T" np "FMT_SZ_T"\n",
				 am, (void *)(uaddr), m->length, m->sglen);

		for (j=0; j<m->sglen; j+=ppp) {

			np = MIN(ppp, m->sglen - j);

			down_read(&mm->mmap_sem);
			rc = get_user_pages(task, mm,
					    uaddr, np, 1, 0, pages, NULL);
			up_read(&mm->mmap_sem);
			if (rc != np)
				goto fail;

			for (n=0; n<np; n++) {
				struct scatterlist *sg = m->sglist[j+n];

				sg->page = pages[n];
				sg->offset = uaddr & (PAGE_SIZE-1);
				sg->length = MIN(m->length - l,
						 (size_t)PAGE_SIZE - sg->offset);
				l += sg->length;
				uaddr += sg->length;
			}
		}
		if (l != m->length)	/* sanity check */
			BUG();

		/* more sanity check */
		page_sanity_check(m, p3validate);
	}
	mmput(mm);
	kfree(pages);
	return PTL_OK;
fail:
	mmput(mm);
	if (DEBUG_P3(p3lib_debug, PTL_DBG_MEMORY))
		p3_print("p3validate: failed on map %p\n", am);

	for (i=0; i<am->maplen; i++) {
		struct p3_ubuf_map *m = am->map[i];

		if (!m->sglist[0]->page)
			break;
		for (j=0; j<m->sglen; j++) {
			if (!m->sglist[j]->page)
				break;
			put_page(m->sglist[j]->page);
			m->sglist[j]->page = NULL;
		}
	}
	kfree(pages);
	return PTL_FAIL;
}
EXPORT_SYMBOL_GPL(p3validate);

void p3invalidate(struct p3_addrmap *am)
{
	size_t i, j;

	for (i=0; i<am->maplen; i++) {
		struct p3_ubuf_map *m = am->map[i];

		if (DEBUG_P3(p3lib_debug, PTL_DBG_MEMORY))
			p3_print("p3invalidate: map %p uaddr %p "
				 "len "FMT_SZ_T" np "FMT_SZ_T"\n",
				 am, (void *)(m->uaddr), m->length, m->sglen);

		for (j=0; j<m->sglen; j++) {
			if (!m->sglist[j]->page)
				continue;
			set_page_dirty_lock(m->sglist[j]->page);
			put_page(m->sglist[j]->page);
			m->sglist[j]->page = NULL;
		}
	}
}
EXPORT_SYMBOL_GPL(p3invalidate);


#endif /* NO_PINNED_PAGES */
