
/* This Cplant(TM) source code is part of the Portals3 Reference
 * Implementation.
 *
 * This Cplant(TM) source code is the property of Sandia National
 * Laboratories.
 *
 * This Cplant(TM) source code is copyrighted by Sandia National
 * Laboratories.
 *
 * The redistribution of this Cplant(TM) source code is subject to the
 * terms of version 2 of the GNU General Public License.
 * (See COPYING, or http://www.gnu.org/licenses/lgpl.html.)
 *
 * Cplant(TM) Copyright 1998-2006 Sandia Corporation. 
 *
 * Under the terms of Contract DE-AC04-94AL85000, there is a non-exclusive
 * license for use of this work by or on behalf of the US Government.
 * Export of this program may require a license from the United States
 * Government.
 */


/* Portals3 is free software; you can redistribute it and/or modify
 * it under the terms of version 2 of the GNU General Public License,
 * as published by the Free Software Foundation.
 *
 * Portals3 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Portals3; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Questions or comments about this library should be sent to:
 *
 * Jim Schutt
 * Sandia National Laboratories, New Mexico
 * P.O. Box 5800
 * Albuquerque, NM 87185-0806
 *
 * jaschut@sandia.gov
 *
 */
#include <p3-config.h>

#include <linux/list.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/err.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/device.h>
#include <linux/cdev.h>
#include <linux/ctype.h>
#include <asm/uaccess.h>
#include <linux/bitops.h>

/* These are all Portals3 include files.
 */
#include "p3utils.h"

#include <p3api/types.h>
#include <p3api/debug.h>

#include <p3/lock.h>
#include <p3/handle.h>
#include <p3/process.h>
#include <p3/forward.h>
#include <p3/nal_types.h>
#include <p3/errno.h>
#include <p3/debug.h>

#include <p3lib/types.h>
#include <p3lib/p3lib.h>
#include <p3lib/nal.h>
#include <p3lib/dispatch.h>
#include <p3lib/p3lib_support.h>

#include <p3/obj_alloc.h>	/* MUST come after p3lib/p3lib.h */

#include "lib-tcpnal.h"
#include "p3_module.h"
#include "p3tcp_module.h"
#include "p3lib_module.h"

/* iface can be a colon-delimited list of up to MAX_P3TCP_IF
 * interface names, e.g., "eth0:eth1:myri0", built from the character
 * set [a-zA-Z0-9_], with a maximum length of MAX_P3TCP_IFNAME.
 *
 * We'll set things up so the first in the list will be assigned to
 * the ptl_interface_t value PTL_NALTYPE_TCP0, and be referenced from
 * the Portals API via the value PTL_IFACE_TCP0, and so on.
 *
 * We'll also set up /sys/class/portals3/p3tcp, and populate it with
 * entries for each interface registered at module insertion time.
 * Entries will be named with a Portals API ptl_interface_t value such
 * as PTL_IFACE_TCP0, and if read will contain the interface name
 * associated with that value.
 */
unsigned p3tcp_if_count;

char p3tcp_if_name[MAX_P3TCP_IF][MAX_P3TCP_IFNAME];

struct p3tcp_kthread p3tcp_kthread[MAX_P3TCP_IF] = {
	{ PTL_NALTYPE_TCP0, },
	{ PTL_NALTYPE_TCP1, },
	{ PTL_NALTYPE_TCP2, },
	{ PTL_NALTYPE_TCP3, }
};

static char *iface = "eth0";
module_param(iface, charp, S_IRUGO);

unsigned int p3tcp_debug, dbg;
module_param(dbg, uint, S_IRUGO | S_IWUSR);

#define P3TCP_VERSION_STRING "Portals Message Passing TCP NAL v"PACKAGE_VERSION

int p3tcp_thread(void *arg)
{
	struct p3tcp_kthread *thread = (struct p3tcp_kthread *)arg;
	struct p3tcp_data *nal_data = p3tcp_data(thread->type);

	daemonize(thread->name);
	allow_signal(SIGSTOP);

	thread->progress_task = current;
	__p3tcp_save_progress_thread(nal_data, thread);

	if (p3tcp_init_private(thread->type, NULL)) {
		printk(KERN_ERR "p3tcp_thread: tcp_lib_start() failed\n");
		set_bit(P3TCP_KTHREAD_ERROR, &thread->flags);
		goto out;
	}
	/* Signal module insert that startup was successful
	 */
	set_bit(P3TCP_KTHREAD_RUNNING, &thread->flags);
	clear_bit(P3TCP_KTHREAD_ERROR, &thread->flags);
	clear_bit(P3TCP_KTHREAD_STOP, &thread->flags);

	up(&thread->sem);

	while(!test_bit(P3TCP_KTHREAD_STOP, &thread->flags))
		p3tcp_chan_poll(nal_data, 1);

	if (DEBUG_P3(p3tcp_debug, PTL_DBG_SHUTDOWN))
		p3_print("p3tcp_thread: received stop signal\n");
            
out:
	p3tcp_libnal_shutdown(nal_data);

	if (DEBUG_P3(p3tcp_debug, PTL_DBG_SHUTDOWN))
		p3_print("p3tcp_thread: signalling shutdown complete\n");

	clear_bit(P3TCP_KTHREAD_RUNNING, &thread->flags);
	up(&thread->sem);

	return 0;
}

void p3tcp_wakeup(struct p3tcp_kthread *thread)
{
	int rc;
	struct task_struct *task = thread->progress_task;

	/* If we weren't called from the progress thread, e.g. for a reply,
	 * send a signal to the progress thread, to interrupt the select
	 * system call out of its infinite wait.
	 */
	if (current == task)
		return;

	rc = force_sig_info(SIGSTOP, (void *)1, task);
	if (rc < 0) {
		p3_print("p3tcp_wakeup: Error signalling progress "
			 "thread: %d\n", -rc);
	}
	else
		wake_up_process(task);
}

static 
void __init p3tcp_parse_modparam(void)
{
	unsigned i = 0;
	char *ptr = iface;

	if (dbg != p3tcp_debug)
		p3tcp_debug = dbg;
	else
		p3tcp_debug = p3lib_debug;

	while (p3tcp_if_count < MAX_P3TCP_IF) {

		while (*ptr && *ptr != ':') {

			if (i >= MAX_P3TCP_IFNAME-1 ||
			    !(isalnum(*ptr) || *ptr == '_')) {
				p3tcp_if_name[p3tcp_if_count][0] = 0;
				i = MAX_P3TCP_IFNAME;
			}
			else
				p3tcp_if_name[p3tcp_if_count][i++] = *ptr;

			ptr++;
		}
		if (p3tcp_if_name[p3tcp_if_count][0]) {
			p3tcp_if_count++;
			i = 0;
		}
		if (!*ptr)
			break;
		ptr++;
	}
}

static struct class_device p3tcp_class_dev = {
	.class = &p3_class,
};

static
ssize_t attr_p3tcp_if0_show(struct class_device *cls, char *buf)
{
	return snprintf(buf, PAGE_SIZE, "%s\n", p3tcp_if_name[0]);
}

static
ssize_t attr_p3tcp_if1_show(struct class_device *cls, char *buf)
{
	return snprintf(buf, PAGE_SIZE, "%s\n", p3tcp_if_name[1]);
}

static
ssize_t attr_p3tcp_if2_show(struct class_device *cls, char *buf)
{
	return snprintf(buf, PAGE_SIZE, "%s\n", p3tcp_if_name[2]);
}

static
ssize_t attr_p3tcp_if3_show(struct class_device *cls, char *buf)
{
	return snprintf(buf, PAGE_SIZE, "%s\n", p3tcp_if_name[3]);
}

static
struct class_device_attribute attr_p3tcp_if[MAX_P3TCP_IF] = {
	{
		.attr = { .name = "PTL_IFACE_TCP0",
			  .owner = THIS_MODULE,
			  .mode = S_IRUGO
		},
		.show = attr_p3tcp_if0_show,
	},
	{
		.attr = { .name = "PTL_IFACE_TCP1",
			  .owner = THIS_MODULE,
			  .mode = S_IRUGO
		},
		.show = attr_p3tcp_if1_show,
	},
	{
		.attr = { .name = "PTL_IFACE_TCP2",
			  .owner = THIS_MODULE,
			  .mode = S_IRUGO
		},
		.show = attr_p3tcp_if2_show,
	},
	{
		.attr = { .name = "PTL_IFACE_TCP3",
			  .owner = THIS_MODULE,
			  .mode = S_IRUGO
		},
		.show = attr_p3tcp_if3_show,
	},
};

static
int __init p3tcp_init(void)
{
	unsigned i;
	int err;

	p3tcp_parse_modparam();

	strlcpy(p3tcp_class_dev.class_id, "p3tcp", BUS_ID_SIZE);
	err = class_device_register(&p3tcp_class_dev);
	if (err) 
		goto fail;

	for (i=0; i<p3tcp_if_count; i++) {

		char name[MAX_P3TCP_IFNAME] = {0,};
		struct p3tcp_kthread *thread = p3tcp_kthread + i;

		err = class_device_create_file(&p3tcp_class_dev,
					       &attr_p3tcp_if[i]);
		if (err)
			goto fail_class_device;

		init_MUTEX_LOCKED(&thread->sem);

		snprintf(name, sizeof(name), "TCP%d", i);
		lib_register_nal(thread->type, name,
				 p3tcp_create_nal, p3tcp_stop_nal, 
				 p3tcp_destroy_nal, p3tcp_pid_ranges);
		snprintf(thread->name, sizeof(thread->name),
			 "p3tcp_%s", p3tcp_if_name[i]);
	}

	for (i=0; i<p3tcp_if_count; i++) {

		long lpid;
		struct p3tcp_kthread *thread = p3tcp_kthread + i;

		lpid = kernel_thread(p3tcp_thread, thread, 0);
		if (lpid < 0) {
			err = lpid;
			goto fail_thread;
		}
		/* Wait until thread signals it is done trying to start,
		 * then check status.
		 */
		down(&p3tcp_kthread[i].sem);

		if (test_bit(P3TCP_KTHREAD_ERROR, &thread->flags) ||
		    !test_bit(P3TCP_KTHREAD_RUNNING, &thread->flags)) {

			err = -EINVAL;
			goto fail_thread;
		}
	}
	printk(KERN_INFO P3TCP_VERSION_STRING"\n");
	return 0;

fail_thread:
	printk(KERN_ERR "p3tcp_init: kernel_thread failed, i = %d\n", i);
	do {
		struct p3tcp_kthread *thread = p3tcp_kthread + i;

		if (test_bit(P3TCP_KTHREAD_RUNNING, &thread->flags)) {
			set_bit(P3TCP_KTHREAD_STOP, &thread->flags);
			down(&thread->sem);
		}
	} while (i-- > 0);

	i = p3tcp_if_count - 1;

fail_class_device:
	do {
		if (p3tcp_kthread[i].name[0]) {
			lib_unregister_nal(p3tcp_kthread[i].type);
			p3tcp_kthread[i].name[0] = 0;
		}
		class_device_remove_file(&p3tcp_class_dev, &attr_p3tcp_if[i]);
		p3tcp_if_name[i][0] = 0;
	} while (i-- > 0);
fail:
	p3tcp_if_count = 0;
	return err;
}

static
void __exit p3tcp_exit(void)
{
	unsigned i = p3tcp_if_count;

	while (i--) {

		struct p3tcp_kthread *thread = p3tcp_kthread + i;

		if (test_bit(P3TCP_KTHREAD_RUNNING, &thread->flags)) {
			if (DEBUG_P3(p3tcp_debug, PTL_DBG_SHUTDOWN))
				p3_print("p3tcp_exit: signalling shutdown "
					 "to %s\n", thread->name);
			set_bit(P3TCP_KTHREAD_STOP, &thread->flags);
			p3tcp_wakeup(thread);
			down(&thread->sem);
		}
		if (thread->name[0]) {
			if (DEBUG_P3(p3tcp_debug, PTL_DBG_SHUTDOWN))
				p3_print("p3tcp_exit: unregistering nal "
					 "type %#x\n", thread->type);
			lib_unregister_nal(thread->type);
			thread->name[0] = 0;
		}
		if (DEBUG_P3(p3tcp_debug, PTL_DBG_SHUTDOWN))
			p3_print("p3tcp_exit: removing sysfs class "
				 "device %s/%s/%s\n",
				 p3_class.name,
				 p3tcp_class_dev.class_id,
				 attr_p3tcp_if[i].attr.name);
		class_device_remove_file(&p3tcp_class_dev, &attr_p3tcp_if[i]);
		p3tcp_if_name[i][0] = 0;
	}
	if (DEBUG_P3(p3tcp_debug, PTL_DBG_SHUTDOWN))
		p3_print("p3tcp_exit: removing sysfs class device %s/%s\n",
			 p3_class.name, p3tcp_class_dev.class_id);
	class_device_unregister(&p3tcp_class_dev);
}

module_init(p3tcp_init);
module_exit(p3tcp_exit);

MODULE_LICENSE("GPL v2");
MODULE_AUTHOR("Jim Schutt");
MODULE_VERSION(PACKAGE_VERSION);
MODULE_DESCRIPTION(P3TCP_VERSION_STRING);
