/*
 * This Cplant(TM) source code is part of the Portals3 Reference
 * Implementation.
 *
 * This Cplant(TM) source code is the property of Sandia National
 * Laboratories.
 *
 * This Cplant(TM) source code is copyrighted by Sandia National
 * Laboratories.
 *
 * The redistribution of this Cplant(TM) source code is subject to the
 * terms of version 2 of the GNU General Public License.
 * (See COPYING, or http://www.gnu.org/licenses/lgpl.html.)
 *
 * Cplant(TM) Copyright 1998-2006 Sandia Corporation. 
 *
 * Under the terms of Contract DE-AC04-94AL85000, there is a non-exclusive
 * license for use of this work by or on behalf of the US Government.
 * Export of this program may require a license from the United States
 * Government.
 */


/* Portals3 is free software; you can redistribute it and/or modify
 * it under the terms of version 2 of the GNU General Public License,
 * as published by the Free Software Foundation.
 *
 * Portals3 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Portals3; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Questions or comments about this library should be sent to:
 *
 * Jim Schutt
 * Sandia National Laboratories, New Mexico
 * P.O. Box 5800
 * Albuquerque, NM 87185-0806
 *
 * jaschut@sandia.gov
 *
 */

#include <unistd.h>
#include <stdio.h>
#include <limits.h>

/* If compiling for user-space (and maybe NIC-space), these need to be
 * the Portals3 versions of the Linux kernel include files, which
 * have been suitably modified for use in user-space code.
 */
#include <linux/list.h>

/* These are all Portals3 include files.
 */
#include <p3-config.h>
#include <p3utils.h>

#include <p3/lock.h>

#include <p3api/types.h>
#include <p3api/nal.h>
#include <p3api/api.h>
#include <p3api/misc.h>

#include <p3/handle.h>
#include <p3/process.h>
#include <p3/obj_alloc.h>
#include <p3/forward.h>
#include <p3/errno.h>

#include "init.h"

int PtlNIInit(ptl_interface_t iface, ptl_pid_t pid,
	      ptl_ni_limits_t *desired, ptl_ni_limits_t *actual,
	      ptl_handle_ni_t *ni_handle)
{
	p3_libreq_t req = PTL_LIBREQ_INIT;
	p3_libres_t res = PTL_LIBRES_INIT;

	ptl_ni_limits_t limits;
	api_ni_t *ni;
	unsigned i, rc = PTL_OK;

	if (!ni_handle) {
		rc = PTL_SEGV;
		goto out;
	}
	if (p3_api_process.init < 0) {
		rc = PTL_NO_INIT;
		goto out;
	}
	/* Look for a duplicate init request.
	 */
	for (i=0; i<PTL_MAX_INTERFACES; i++) {
		ni = p3_api_process.ni[i];
		if (ni && ni->nal.type == iface) {
			*ni_handle = PTL_OBJ_HNDL(ni);
			goto success;
		}
	}
	if (!(ni = p3_malloc(sizeof(*ni)))) {
		rc = PTL_NO_SPACE;
		goto out;
	}
	p3lock_init(&ni->obj_alloc);

	/* FIXME: we could add a call here to a NAL API-side function to
	 * put something NAL-specific in req.t.niinit.data.
	 */
	req.call_index = P3LIB_NI_INIT;
	req.value = pid;

	if (desired) limits = *desired;
	else {
		/* We'll put infinite values here to make sure the NAL
		 * implementation does its job correctly on the library side.
		 */
		limits.max_mes = INT_MAX;
		limits.max_mds = INT_MAX;
		limits.max_eqs = INT_MAX;
		limits.max_ac_index = INT_MAX;
		limits.max_pt_index = INT_MAX;
		limits.max_md_iovecs = INT_MAX;
		limits.max_me_list = INT_MAX;
		limits.max_getput_md = INT_MAX;
	}
	req.t.niinit.limits = limits;
	req.t.niinit.type = iface;

	p3_forward(&req, sizeof(req), &res, sizeof(res));
	if ((rc = res.status) != PTL_OK) {
		p3_free(ni);
		goto out;
	}
	p3_api_process.ni[PTL_NI_INDEX(res.handle)] = ni;
	*ni_handle = res.handle;

	ni->id = *ni_handle;
	ni->limits = res.t.niinit.limits;

	INIT_LIST_HEAD(&ni->free_eq);
	PTL_ALLOC_INIT_TBL(&ni->eq);

	ni->nal.type = iface;
	ni->nal.private = NULL;		/* FIXME */
	ni->nal.errstr = NULL;		/* FIXME */
success:
	p3_api_process.init++;
	if (actual)
		*actual = ni->limits;

	/* FIXME:  Hmmmmm, does the API really need locking?
	 */
	p3_lock(&ni->obj_alloc);
	SET_OBJ(ni, OBJ_INUSE);
	p3_unlock(&ni->obj_alloc);
out:
	return rc;
}

int PtlNIEqDump(ptl_handle_ni_t ni_handle)
{
	api_ni_t *ni;
	unsigned if_idx = PTL_NI_INDEX(ni_handle);
	unsigned i;

	if (p3_api_process.init < 0)
		return PTL_NO_INIT;

	if (!(if_idx < PTL_MAX_INTERFACES && (ni = p3_api_process.ni[if_idx])))
		return PTL_NI_INVALID;

	printf("%d: ni eq tbl %p rows %d/%d\n",
	       getpid(), ni->eq.tbl, ni->eq.next_row, ni->eq.num_rows);

	if (!ni->eq.tbl)
		return PTL_OK;

	for (i=0; i<ni->eq.num_rows; i++)
		printf("%d: ni eq tbl row %p\n", getpid(), ni->eq.tbl[i]);

	return PTL_OK;
}

void __free_all_eq(api_ni_t *ni)
{
	unsigned i, j;

	/* Look for any event queues that are in use, and free them.
	 * FIXME: obj_alloc.h needs to have an iterator method for
	 * this type of thing.
	 */
	for (i=0; i<ni->eq.next_row; i++)
		for (j=0; j<PTL_INDX_MAX_COL; j++) {
			api_eq_t *eq = ni->eq.tbl[i] + j;
			if (TST_OBJ(eq, OBJ_INUSE)) {
				p3_free(eq->base);
				ptl_obj_free(eq, ni);
			}
		}
}

int PtlNIFini(ptl_handle_ni_t ni_handle)
{
	p3_libreq_t req = PTL_LIBREQ_INIT;
	p3_libres_t res = PTL_LIBRES_INIT;
	api_ni_t *ni;
	unsigned if_idx = PTL_NI_INDEX(ni_handle);

	if (p3_api_process.init < 0)
		return PTL_NO_INIT;

	if (!(if_idx < PTL_MAX_INTERFACES && (ni = p3_api_process.ni[if_idx])))
		return PTL_NI_INVALID;

	p3_api_process.ni[if_idx] = NULL;
	p3_api_process.init--;

	req.call_index = P3LIB_NI_FINI;
	req.handle = ni_handle;

	p3_forward(&req, sizeof(req), &res, sizeof(res));

	__free_all_eq(ni);
	PTL_ALLOC_FREE_TBL(eq, ni);
	p3_free(ni);

	return PTL_OK;
}

int PtlNIStatus(ptl_handle_ni_t ni_handle,
		ptl_sr_index_t register_index,
		ptl_sr_value_t *status)
{
	p3_libreq_t req = PTL_LIBREQ_INIT;
	p3_libres_t res = PTL_LIBRES_INIT;

	if (!status)
		return PTL_SEGV;

	if (p3_api_process.init < 0)
		return PTL_NO_INIT;

	if (!(PTL_NI_INDEX(ni_handle) < PTL_MAX_INTERFACES &&
	      p3_api_process.ni[PTL_NI_INDEX(ni_handle)]))
		return PTL_NI_INVALID;

	req.call_index = P3LIB_NI_STATUS;
	req.handle = ni_handle;
	req.value = register_index;

	p3_forward(&req, sizeof(req), &res, sizeof(res));

	*status = res.value;
	return res.status;
}

int PtlNIDist(ptl_handle_ni_t ni_handle,
	      ptl_process_id_t process,
	      unsigned long *distance)
{
	p3_libreq_t req = PTL_LIBREQ_INIT;
	p3_libres_t res = PTL_LIBRES_INIT;

	if (!distance)
		return PTL_SEGV;

	if (p3_api_process.init < 0)
		return PTL_NO_INIT;

	if (!(PTL_NI_INDEX(ni_handle) < PTL_MAX_INTERFACES &&
	      p3_api_process.ni[PTL_NI_INDEX(ni_handle)]))
		return PTL_NI_INVALID;

	req.call_index = P3LIB_NI_DIST;
	req.handle = ni_handle;
	req.value = process.nid;

	p3_forward(&req, sizeof(req), &res, sizeof(res));

	*distance = res.value;
	return res.status;
}

int PtlNIHandle(ptl_handle_any_t handle, ptl_handle_ni_t *ni)
{
	if (!ni)
		return PTL_SEGV;

	if (p3_api_process.init < 0)
		return PTL_NO_INIT;

	if (!(PTL_NI_INDEX(handle) < PTL_MAX_INTERFACES &&
	      p3_api_process.ni[PTL_NI_INDEX(handle)]))
		return PTL_HANDLE_INVALID;

	*ni = PTL_NI_HNDL(handle);
	return PTL_OK;
}

/*
 * This function isn't part of the spec.  See p3api/debug.h for an explanation.
 */
int PtlTblDump(ptl_handle_ni_t ni_handle, int pt_index) 
{
	p3_libreq_t req = PTL_LIBREQ_INIT;
	p3_libres_t res = PTL_LIBRES_INIT;

	if (p3_api_process.init < 0)
		return PTL_NO_INIT;

	if (!(PTL_NI_INDEX(ni_handle) < PTL_MAX_INTERFACES &&
	      p3_api_process.ni[PTL_NI_INDEX(ni_handle)]))
		return PTL_NI_INVALID;

	req.call_index = P3LIB_TBLDUMP;
	req.handle = ni_handle;
	req.value = pt_index;

	p3_forward(&req, sizeof(req), &res, sizeof(res));

	return res.status;
}

/*
 * This function isn't part of the spec either.  See p3api/debug.h for an
 * explanation.
 */ 
unsigned int PtlNIDebug(ptl_handle_ni_t ni_handle, unsigned int mask)
{
	p3_libreq_t req = PTL_LIBREQ_INIT;
	p3_libres_t res = PTL_LIBRES_INIT;

	if (p3_api_process.init < 0)
		return 0;

	if (!(ni_handle == PTL_INVALID_HANDLE) &&
	    !(PTL_NI_INDEX(ni_handle) < PTL_MAX_INTERFACES &&
	      p3_api_process.ni[PTL_NI_INDEX(ni_handle)]))
		return 0;

	req.call_index = P3LIB_NI_DEBUG;
	req.handle = ni_handle;
	req.value = mask;

	p3_forward(&req, sizeof(req), &res, sizeof(res));

	return res.value;
}
